-- circular.e
include myLibs/myDebug.e

constant
  CIRC_SIZE=1, CIRC_WRITE_PTR=2, CIRC_READ_PTR=3, CIRC_BUFFER = 4

sequence circularBuffers
circularBuffers = {}
  
--------------------------------------------------------------------------------

global function circNew(integer size)
  circularBuffers = append(circularBuffers, {size, 1, 1, repeat(0, size)})
  return length(circularBuffers)
end function

--------------------------------------------------------------------------------

global procedure circDestroy(integer circ)
  circularBuffers[circ] = {}
end procedure

--------------------------------------------------------------------------------

global procedure circClear(integer circ)
  circularBuffers[circ][CIRC_READ_PTR] = circularBuffers[circ][CIRC_WRITE_PTR]
end procedure

--------------------------------------------------------------------------------

function circPrevious(integer circ, integer ptr)
  circularBuffers[circ][ptr] -= 1
  if circularBuffers[circ][ptr] < 1 then circularBuffers[circ][ptr] = circularBuffers[circ][CIRC_SIZE] end if
  return circularBuffers[circ][ptr]
end function

--------------------------------------------------------------------------------

function circNext(integer circ, integer ptr)
  circularBuffers[circ][ptr] += 1
  if circularBuffers[circ][ptr] > circularBuffers[circ][CIRC_SIZE] then
    circularBuffers[circ][ptr] = 1 end if
  return circularBuffers[circ][ptr]
end function

--------------------------------------------------------------------------------

global function circReadIndex(integer circ)
  return circularBuffers[circ][CIRC_READ_PTR]
end function

--------------------------------------------------------------------------------

global function circWriteIndex(integer circ)
  return circularBuffers[circ][CIRC_WRITE_PTR]
end function

--------------------------------------------------------------------------------

global procedure circWriteAt(integer circ, integer ptr, object o)
  if ptr = 0 then return end if
  circularBuffers[circ][CIRC_BUFFER][ptr] = o
end procedure

--------------------------------------------------------------------------------

global function circReadAt(integer circ, integer ptr)
  if ptr = 0 then return {} end if
  return circularBuffers[circ][CIRC_BUFFER][ptr]
end function

--------------------------------------------------------------------------------

-- write object o at current write pointer then increase write pointer
-- overflow_allowed should be 0 by default
global procedure circWrite(integer circ, object o, integer overflow_allowed)
  integer ptr

--  logMsg(sprintf("circWrite(%d, %s)", { circ, o}))
  ptr = circularBuffers[circ][CIRC_WRITE_PTR]
  if ptr = circularBuffers[circ][CIRC_READ_PTR]-1 then
    logMsg(sprintf("circWrite(%d): buffer overflow", circ))
    if not overflow_allowed then return end if
  end if
--  logMsg(sprintf("circWriteAt(%d, %d, %s)", { circ, ptr, o}))
  circWriteAt(circ, ptr, o)
  circularBuffers[circ][CIRC_WRITE_PTR] = circNext(circ, CIRC_WRITE_PTR)
end procedure

--------------------------------------------------------------------------------

-- read object at current read pointer then increase read pointer
global function circRead(integer circ)
  integer ptr
  object o

  ptr = circularBuffers[circ][CIRC_READ_PTR]
  if ptr = 0 then return {} end if
  if ptr = circularBuffers[circ][CIRC_WRITE_PTR] then
    logMsg(sprintf("circRead(%d): no data to read", circ))
    return {}
  end if
  o = circReadAt(circ, ptr)
  circularBuffers[circ][CIRC_READ_PTR] = circNext(circ, CIRC_READ_PTR)
  return o
end function

--------------------------------------------------------------------------------

global function circAvailable(integer circ)
  if circularBuffers[circ][CIRC_WRITE_PTR] >= circularBuffers[circ][CIRC_READ_PTR] then
    return circularBuffers[circ][CIRC_WRITE_PTR] - circularBuffers[circ][CIRC_READ_PTR]
  else    
    return (circularBuffers[circ][CIRC_SIZE] - circularBuffers[circ][CIRC_READ_PTR]) + circularBuffers[circ][CIRC_WRITE_PTR]
  end if
end function

--------------------------------------------------------------------------------

global function circNewData(integer circ)
  integer write_ptr, read_ptr
  sequence result

  write_ptr = circularBuffers[circ][CIRC_WRITE_PTR]
  read_ptr = circularBuffers[circ][CIRC_READ_PTR]
  result = {}
  -- printf(f_debug, "write_ptr = %d, read_ptr = %d\n", {write_ptr, read_ptr})
  if write_ptr = read_ptr then
    return {}
  elsif write_ptr > read_ptr then
    result = circularBuffers[circ][CIRC_BUFFER][ read_ptr..write_ptr-1 ]
  else    
    result = circularBuffers[circ][CIRC_BUFFER][ read_ptr..circularBuffers[circ][CIRC_SIZE] ] &
           circularBuffers[circ][CIRC_BUFFER][ 1..write_ptr-1 ]
  end if
--  circClear(circ)
  return result
end function
