-- common.e
-- library of common reusable functions
-- v1.11
-- 2010-11-10
-- history:
-- v1.0   2010-07-26
-- v1.01
--   changed french terms to english ones
--   saveCSV uses quotes around strings
-- v1.02
--   changed id from atom to object in findID
-- v1.03  2010-07-28
--   added dateStamp function
-- v1.04  2010-07-29
--   added min and max functions
-- v1.05  2010-08-04
--   added  function appendIfNew
-- v1.06  2010-09-29
--   added function utf8_to_ascii
-- v1.07  2010-10-04
--   new INI files management
-- v1.08  2010-11-05
--   added searchINI
-- v1.09  2010-11-05
-- added insert, delete, replace, mid from Irv Mullins (irv@ellijay.com)
-- v1.10  2010-11-06
-- modified strtok to convert floats
-- v1.11  2010-11-10
-- added copyFile

include get.e
include md5.e
include msgbox.e
include file.e
include myTypes.e
include mySeq.e
include myDebug.e
include idle.e
--with trace

global sequence SLASH
  if (platform() = WIN32) or
     (platform() = DOS32) then
    SLASH = "\\"
  else
    SLASH = "/"
  end if

------------------------------------------------------------------------------

global function itemPos(sequence sht, integer fld, object val, integer start)
  if fld = 0 then return {} end if
  for i = start to length(sht) do
    if compare(sht[i][fld], val) = 0 then
      return i
    end if
  end for
  return 0
end function

------------------------------------------------------------------------------

global function commonField(sequence sht, integer fld, object val, integer start)
-- lists records from sheet sht with value val in field fld
  sequence res

  if fld = 0 then return {} end if
  res = {}
  for i = start to length(sht) do
    if equal(sht[i][fld], val) then
      res = append(res, i)
    end if
  end for
  return res
end function

------------------------------------------------------------------------------

global function findID(sequence sht, object id, integer start)
-- find the record in sheet sht which ID (value in first column) is id
-- returns the record (sequence) if successful
-- returns 0 (atom) if not
  sequence s
  
--  if atom(id) and (id = 0) then return 0 end if
  if sequence(id) and (length(id) = 0) then return 0 end if
  s = commonField(sht, 1, id, start)
  if length(s) = 1 then
    return sht[s[1]]
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

global function findRecNumByField(sequence sht, integer fld, object id, integer start)
-- find the record in sheet sht which ID (value in first column) is id
-- returns the record number if successful
-- returns 0 (atom) if not
  sequence s

  s = commonField(sht, fld, id, start)
--  printf(f_debug, "s = %s\n", {sequenceDump(s)})
  if length(s) then
    return s[1]
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

global function findRecordByField(sequence sht, integer fld, object id, integer start)
-- find the record in sheet sht which ID (value in first column) is id
-- returns the record (sequence) if successful
-- returns 0 (atom) if not
  integer i

  i = findRecNumByField(sht, fld, id, start)
  if i > 0 then
    return sht[i]
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

global function vlookup(object id, sequence sht, integer searchField,
                        integer targetField, integer start, object default)
  sequence s

  s = findRecordByField(sht, searchField, id, start)
  if sequence(s) then
    return s[targetField]
  else
    return default
  end if
end function

------------------------------------------------------------------------------

global function selectedFields(sequence sht, integer fld, object val, sequence fields, integer start)
-- list selected fields from sheet sht records with value val in field fld
-- ex: selectedFields(tab, 2, "John", {1,3}) will list items having
-- "John" in second column and return a sequence containing columns 1 and 3
--NB is this still useful? commonField returns the whole records
  sequence s, res
  integer l

  if (fld = 0) or (length(fields) = 0) then return {} end if
  s = commonField(sht, fld, val, start)
  l = length(s)
  res = {}
  if l = 1 then
    for i = 1 to length(fields) do
      res = append(res, sht[s[1]][fields[i]])
    end for
    return res
  else
    return {}
  end if
end function

------------------------------------------------------------------------------

global function commonFields(sequence sht, sequence fields, sequence vals, integer start)
-- list records from sheet sht with selected values in selected fields
-- ex: commonFields(tab, {1,3}, {5,"John"}) will list items having 5 in first column and "John" in third
  sequence res
  integer found
  
  if (length(fields) = 0) or (length(vals) = 0) or (length(fields) != length(vals)) then return {} end if
  res = {}
  for i = start to length(sht) do
    found = 1
    for j = 1 to length(fields) do
      if compare(sht[i][fields[j]], vals[j]) != 0 then
        found = 0
      end if
    end for
    if found = 1 then  
      res = append(res, i)
    end if
  end for
  return res
end function


