include myDebug.e

global constant
  STACK_NAME = 1, STACK_BUFFER = 2

global sequence LiFos
LiFos = {}

global integer rtn_ident_lifo

-------------------------------------------------------------------------------

function identifyStack(sequence path, integer level, integer n)
--  printf(f_debug, "  path: '%s', level: %d, n: %d\n", {path, level, n})
  if level = 1 then
    if    n =  1 then return "STACK_NAME"
    elsif n =  2 then return "STACK_BUFFER"
    end if
  end if
  return sprintf("%d",n)
end function
rtn_ident_lifo = routine_id("identifyStack")

------------------------------------------------------------------------------

global function stackCheck(integer stack_id)
  integer lg

  lg = length(LiFos)
  if lg = 0 then
    logMsg("no LiFo stack_id available")
    return 0
  end if
  if (stack_id < 1) or (stack_id > lg) then
    logMsg(sprintf("no stack_id %d available", stack_id))
    return 0
  end if
  return 1
end function

-----------------------------------------------------------------------------

global function stackPush(integer stack_id, object o)
  sequence buffer

--  logMsg(sprintf("stackPush(%d, '%s')", {stack_id, sprint(o)}))
  if not stackCheck(stack_id) then return 0 end if
  buffer = LiFos[stack_id][STACK_BUFFER]
  LiFos[stack_id][STACK_BUFFER] = prepend(buffer, o)
  return 1
end function

------------------------------------------------------------------------------

global function stackLast(integer stack_id, object notFound)
  sequence buffer

--  logMsg(sprintf("stackLast(%d)", {stack_id}))
  buffer = LiFos[stack_id][STACK_BUFFER]
--  analyzeObject(buffer, sprintf("stackLast(%d)", {stack_id}), f_debug, 0)
  if length(buffer) then
    return buffer[1]
  else
    return notFound
  end if
end function

-----------------------------------------------------------------------------

procedure remove_last(integer stack_id)
  sequence buffer

  buffer = LiFos[stack_id][STACK_BUFFER]
  if length(buffer) > 1 then
    buffer = buffer[2..$]
  else
    buffer = {}
  end if
  LiFos[stack_id][STACK_BUFFER] = buffer
end procedure

-----------------------------------------------------------------------------

global function stackPop(integer stack_id, object notFound)
  object x

--  logMsg(sprintf("stackPop(%d)", {stack_id}))
  if not stackCheck(stack_id) then return notFound end if
  x = stackLast(stack_id, notFound)
--  analyzeObject(x, sprintf("stackPop(%d)", {stack_id}), f_debug, 0)
  remove_last(stack_id)
  return x
end function

------------------------------------------------------------------------------

global function stackAt(integer stack_id, integer i, object notFound)
  sequence buffer

--  logMsg(sprintf("stackAt(%d, %d)", {stack_id, i}))
  if not stackCheck(stack_id) then return notFound end if
  buffer = LiFos[stack_id][STACK_BUFFER]
  if length(buffer) < i then
    return notFound
  else
    return buffer[i]
  end if
end function

-----------------------------------------------------------------------------

global function stackPath(integer stack_id)
  sequence s
  sequence buffer

  logMsg(sprintf("stackPath(%d)", {stack_id}))
  buffer = LiFos[stack_id][STACK_BUFFER]
  s = ""
  for i = 1 to length(buffer) do
    s &= buffer[i]
  end for
  return s
end function

--------------------------------------------------------------------------------

global function stackFind(sequence stack_name)
  return itemPos(LiFos, STACK_NAME, stack_name, 1)
end function

--------------------------------------------------------------------------------

global function stackSize(integer stack_id)
  return length(LiFos[stack_id][STACK_BUFFER])
end function

--------------------------------------------------------------------------------

global procedure stackClear(integer stack_id)
--  logMsg(sprintf("stackClear(%d)", {stack_id}))
  if length(LiFos)=0 then
    logMsg("stackClear: no stack_id available")
    return
  end if
  LiFos[stack_id][STACK_BUFFER] = {}
end procedure

--------------------------------------------------------------------------------

global function stackNew(sequence stack_name)
  integer stack_id

--  logMsg(sprintf("stackNew(%s)", {stack_name}))
  stack_id = stackFind(stack_name)
  if stack_id then
    stackClear(stack_id)
  else
    LiFos = append(LiFos, {stack_name, {}})
    stack_id = length(LiFos)
  end if
  return stack_id
end function

