-- Program then    TNCNX.PAS
-- Object then     Delphi component which implement the TCP/IP telnet protocol
--         including some options negociations.
--         RFC854, RFC885, RFC779, RFC1091
-- Author then     Franois PIETTE
-- EMail then    http then//users.swing.be/francois.piette  francois.piette@swing.be
--         http then//www.rtfm.be/fpiette       francois.piette@rtfm.be
--         francois.piette@pophost.eunet.be
-- Creation then   April, 1996
-- Version then    2.09
-- Support then    Use the mailing list twsocket@rtfm.be See website for details.
-- Legal issues then Copyright (C) 1996-2002 by Franois PIETTE
--         Rue de Grady 24, 4053 Embourg, Belgium. Fax then +32-4-365.74.56
--         <francois.piette@pophost.eunet.be>
--
--         This software is provided "as-is", without any express or
--         implied warranty.  In no event will the author be held liable
--         for any  damages arising from the use of this software.
--
--         Permission is granted to anyone to use this software for any
--         purpose, including commercial applications, and to alter it
--         and redistribute it freely, subject to the following
--         restrictions then
--
--         1. The origin of this software must not be misrepresented,
--          you must not claim that you wrote the original software.
--          If you use this software in a product, an acknowledgment
--          in the product documentation would be appreciated but is
--          not required.
--
--         2. Altered source versions must be plainly marked as such, and
--          must not be misrepresented as being the original software.
--
--         3. This notice may not be removed or altered from any source
--          distribution.
--
--         4. You must register this software by sending a picture postcard
--          to the author. Use a nice stamp and mention your name, street
--          address, EMail address and any comment you like to say.
--
-- Updates then
-- Jul 22, 1997 Adapted to Delphi 3
-- Sep 5, 1997  Added version information, removed old code, added OnTermType
--        Renamed some indentifier to be more standard.
-- Sep 24, 1997 V2.03 Added procedures to negociate options
-- May 12, 1998 V2.04 Changed NegociateOption to properly handle unwanted
--        option as Jan Tomasek <xtomasej@feld.cvut.cz> suggested.
-- Aug 10, 1998 V2.05 Cleared activeConnections[cnx][SUB_OPTION] after NegociateSubOption as Jan
--        Tomasek <xtomasej@feld.cvut.cz> suggested.
-- Aug 15, 1999 V2.06 Moved Notification procedure to global section for
--        BCB4 compatibility
-- Aug 20, 1999 V2.07 Added compile time options. Revised for BCB4.
-- Jun 18, 2001 V2.08 Use AllocateHWnd and DeallocateHWnd from wsocket.
-- Oct 23, 2002 V2.09 Changed Buffer arg in OnDataAvailable to Pointer instead
--        of PChar to avoid Delphi 7 messing everything with AnsiChar.
-- Oct 23, 2015 V2.09 Ported to OpenEuphoria 4 by Jean-Marc DURO
-- Mar 31, 2016 V2.09 Ported to Euphoria 3.11 by Jean-Marc DURO

include dll.e
include eunet.e
include Win32Lib.ew  -- doEvents
include myDebug.e
include myConv.e
include mySeq.e as seq
include myFile.e
include common.e
include buffer.e

constant
--  TnCnxVersion     = 209,
--  CopyRight = " TTnCnx (c) 1996-2002 F. Piette V2.09 ",
  FLocation = "TNCNX",
  FTermType = "VT100",

  -- Telnet command characters                       
  TNCH_EOR        = 239,  -- #EF End Of Record (preceded by IAC)     
  TNCH_SE         = 240,  -- #F0 End of subnegociation parameters    
--  TNCH_NOP        = 241,  -- #F1 No operation
--  TNCH_DATA_MARK  = 242,  -- #F2 Data stream portion of a Synch
--  TNCH_BREAK      = 243,  -- #F3 NVT charcater break
  TNCH_IP         = 244,  -- #F4 Interrupt process           
  TNCH_AO         = 245,  -- #F5 Abort output              
  TNCH_AYT        = 246,  -- #F6 Are you there             
  TNCH_EC         = 247,  -- #F7 Erase character             
  TNCH_EL         = 248,  -- #F8 Erase line              
--  TNCH_GA         = 249,  -- #F9 Go ahead
  TNCH_SB         = 250,  -- #FA Subnegociation            
  TNCH_WILL       = 251,  -- #FB Will                  
  TNCH_WONT       = 252,  -- #FC Wont                  
  TNCH_DO         = 253,  -- #FD Do                  
  TNCH_DONT       = 254,  -- #FE Dont                  
  TNCH_IAC        = 255,  -- #FF IAC                   

  -- Telnet options                            
  TN_TRANSMIT_BINARY = 0,   -- #00 
  TN_ECHO            = 1,   -- #01 
--  TN_RECONNECTION    = 2,   -- #02
  TN_SUPPRESS_GA     = 3,   -- #03 
--  TN_MSG_SZ_NEGOC    = 4,   -- #04
--  TN_STATUS          = 5,   -- #05
--  TN_TIMING_MARK     = 6,   -- #06
--  TN_NOPTIONS        = 6,   -- #06
--  TN_DET             = 20,  -- #14
  TN_SEND_LOC        = 23,  -- #17 
  TN_TERMTYPE        = 24,  -- #18 
  TN_EOR             = 25,  -- #19 
--  TN_NAWS            = 31,  -- #1F
--  TN_TERMSPEED       = 32,  -- #20
--  TN_TFC             = 33,  -- #21
--  TN_XDISPLOC        = 35,  -- #23
--  TN_EXOPL           = 255, -- #FF

  TN_TTYPE_SEND      = 1,
  TN_TTYPE_IS        = 0

global constant
  TCP_HOST = 1, TCP_PORT = 2, TCP_SOCKET = 3, TCP_STATUS = 4,
  LAST_IO = 5, BUFFER = 6, IAC = 7, VERB = 8, SUB_OPTION = 9,
  SUB_NEGOC = 10


global constant DISCONNECTED = 0, CONNECTED = 1

global sequence activeConnections, activeSockets
activeConnections = {}
activeSockets = {}

global integer onTheFlowDisplayRoutine
onTheFlowDisplayRoutine = 0

global integer rtn_ident_cnx

object void
void = 0

constant maxTime = 30

integer
  RemoteBinMode,
  LocalBinMode,
  FLocalEcho,
  spga,
  FTType

integer displayBuffer
displayBuffer = bufferNew()

-------------------------------------------------------------------------------

function identifyConnection(sequence path, integer level, integer n)
--  printf(f_debug, "  path: '%s', level: %d, n: %d\n", {path, level, n})
  if level = 1 then
    if    n =  1 then return "TCP_HOST"
    elsif n =  2 then return "TCP_PORT"
    elsif n =  3 then return "TCP_SOCKET"
    elsif n =  4 then return "TCP_STATUS"
    elsif n =  5 then return "LAST_IO"
    elsif n =  6 then return "BUFFER"
    elsif n =  7 then return "IAC"
    elsif n =  8 then return "VERB"
    elsif n =  9 then return "SUB_OPTION"
    elsif n = 10 then return "SUB_NEGOC"
    end if
  end if
  return sprintf("%d",n)
end function
rtn_ident_cnx = routine_id("identifyConnection")

------------------------------------------------------------------------------

global procedure handleError()
  sequence s

  s = eunet_get_error()
  void  = message_box(s[2], sprintf("Telnet Error %d", s[1]), MB_OK)
end procedure

------------------------------------------------------------------------------

global function checkLink(integer cnx)
  sequence host
  integer port, lg
  atom tcpSocket

  lg = length(activeConnections)
  if lg = 0 then
    logMsg("no active connection available")
    return 0
  end if
  if (cnx < 1) or (cnx > lg) then
    logMsg(sprintf("no connection %d available", cnx))
    return 0
  end if
  host = activeConnections[cnx][TCP_HOST]
  port = activeConnections[cnx][TCP_PORT]
  if activeConnections[cnx][TCP_STATUS] = DISCONNECTED then
    logMsg(sprintf("not connected to host %s:%d", {host, port}))
    return 0
  end if
  tcpSocket = activeConnections[cnx][TCP_SOCKET]
  if tcpSocket = 0 then
    logMsg(sprintf("no socket to host %s:%d", {host, port}))
    return 0
  end if
  return 1
end function

------------------------------------------------------------------------------

global function createListeningSocket(integer Port)
  sequence Address
  atom hSockListen

  -- create a TCP listening socket
  hSockListen = eunet_new_socket(AF_INET, SOCK_STREAM, 0)
  -- Setup an IP address/port structure
  Address = sprintf("0.0.0.0:%d", Port)
  -- Bind this end of the socket to the local port
  if eunet_bind(hSockListen, Address) < 0 then
    logMsg(sprintf("Binding to 0.0.0.0:%d has failed", Port))
    handleError()
    return 0
  end if
  -- Listen for a connection request
  if eunet_listen(hSockListen, 0) < 0 then
    handleError()
    return 0
  end if
  logMsg(sprintf("Waiting for connections on port %d", Port))
  return hSockListen
end function

------------------------------------------------------------------------------

global procedure sockDisconnect (atom tcpSocket)
-- Close the connection if one already exists.
  -- Disconnect the existing socket
  void = eunet_close_socket(tcpSocket)
end procedure

------------------------------------------------------------------------------

global function sockConnect (sequence Host, integer Port)
  object Address
  atom tcpSocket

  -- Create a stream (TCP) socket
  tcpSocket = eunet_new_socket(AF_INET, SOCK_STREAM, 0)
  if tcpSocket <= 0 then
    handleError()
  end if
  -- Setup an IP address/port structure
  -- Use the remote address and port specified in the edit boxes
  Address = sprintf("%s:%d", {Host, Port})
  -- Connect the new socket to the remote host
  if eunet_connect(tcpSocket, Address) != 0 then
    handleError()
    void = eunet_close_socket(tcpSocket)
    return 0
  end if

  return tcpSocket
end function

--------------------------------------------------------------------------------

global function telnetSend(integer cnx, sequence cmd)
  integer  status
  
  logMsg(sprintf("telnetSend(%d, '%s')", {cnx, cmd}))
  if not checkLink(cnx) then return 0 end if
  status = eunet_send(activeConnections[cnx][TCP_SOCKET], cmd, 0)
  if status > 0 then
    logMsg(sprintf("telnetSend: Sent '%s' %s (%d bytes) on connection %d", {showPrintable(cmd), sprint(cmd), status, cnx}))
    activeConnections[cnx][LAST_IO] = time()
  else
    handleError()
    logMsg(sprintf("telnetSend: Error while sending '%s' on connection %d", {cmd, cnx}))
  end if
  return status
end function

--------------------------------------------------------------------------------

procedure Answer(integer cnx, integer chAns, integer chOption)
  sequence Buf

  Buf = {TNCH_IAC, chAns, chOption}
  void = telnetSend(cnx, Buf)
end procedure

--------------------------------------------------------------------------------

procedure NegociateSubOption(integer cnx, sequence strSubOption)
  sequence Buf

  if strSubOption[1] = TN_TERMTYPE then
    if strSubOption[2] = TN_TTYPE_SEND then
      Buf = {TNCH_IAC, TNCH_SB, TN_TERMTYPE, TN_TTYPE_IS} & FTermType & {TNCH_IAC, TNCH_SE}
      void = telnetSend(cnx, Buf)
    end if
  else
--    puts(f_debug, "Unknown suboption\r\n"); 
  end if
end procedure

--------------------------------------------------------------------------------

procedure NegociateOption(integer cnx, integer chAction, integer chOption)
  sequence Buf
  
  if chOption = TN_TRANSMIT_BINARY then
    if chAction = TNCH_WILL then
      Answer(cnx, TNCH_DO, chOption)
      RemoteBinMode = 1
      LocalBinMode  = 1
    elsif chAction = TNCH_WONT then
      if RemoteBinMode then
        RemoteBinMode = 0
        LocalBinMode  = 0
      end if
    end if
  elsif chOption = TN_ECHO then
    if chAction = TNCH_WILL then
      Answer(cnx, TNCH_DO, chOption)
      FLocalEcho = 0
    elsif chAction = TNCH_WONT then
      FLocalEcho = 1
    end if
  elsif chOption = TN_SUPPRESS_GA then
    if chAction = TNCH_WILL then
      Answer(cnx, TNCH_DO, chOption)
      spga = 1
    end if
  elsif chOption = TN_TERMTYPE then
    if chAction = TNCH_DO then
      Answer(cnx, TNCH_WILL, chOption)
      FTType = 1
    end if
  elsif chOption = TN_SEND_LOC then
    if chAction = TNCH_DO then
      Answer(cnx, TNCH_WILL, chOption)
      Buf = {TNCH_IAC, TNCH_SB, TN_SEND_LOC, FLocation, TNCH_IAC, TNCH_SE}
      void = telnetSend(cnx, Buf)
    end if
  elsif chOption = TN_EOR then
    if chAction = TNCH_DO then
      Answer(cnx, TNCH_WILL, chOption)
      FTType = 1
    end if
  else
--    Answer(cnx, TNCH_WONT, chOption);
    -- Jan Tomasek <xtomasej@feld.cvut.cz>
    if chAction = TNCH_WILL then
      Answer(cnx, TNCH_DONT, chOption)
    else
      Answer(cnx, TNCH_WONT, chOption)
    end if
  end if
end procedure

--------------------------------------------------------------------------------

procedure AddChar(integer cnx, integer Ch)
  bufferWrite(displayBuffer, {Ch})
  bufferWrite(activeConnections[cnx][BUFFER], {Ch})
end procedure

--------------------------------------------------------------------------------

procedure ReceiveChar(integer cnx, integer Ch)

  if activeConnections[cnx][VERB] != #0 then
    NegociateOption(cnx, activeConnections[cnx][VERB], Ch)
    activeConnections[cnx][VERB]     = #0
    activeConnections[cnx][SUB_OPTION] = ""
    return
  end if

  if activeConnections[cnx][SUB_NEGOC] then
    if Ch = TNCH_SE then
      activeConnections[cnx][SUB_NEGOC]  = 0
      NegociateSubOption(cnx, activeConnections[cnx][SUB_OPTION])
      activeConnections[cnx][SUB_OPTION] = ""
    else
      activeConnections[cnx][SUB_OPTION] = activeConnections[cnx][SUB_OPTION] & Ch
    end if
    return
  end if

  if activeConnections[cnx][IAC] then
    if Ch = TNCH_IAC then
      AddChar(cnx, Ch)
      activeConnections[cnx][IAC] = 0
    elsif find(Ch, {TNCH_DO, TNCH_WILL, TNCH_DONT, TNCH_WONT}) then
      activeConnections[cnx][IAC]   = 0
      activeConnections[cnx][VERB] = Ch
    elsif Ch = TNCH_EOR then
      puts(f_debug, "TNCH_EOR\r\n")
      activeConnections[cnx][IAC]   = 0
    elsif Ch = TNCH_SB then
--      puts(f_debug, "Subnegociation\r\n");
      activeConnections[cnx][SUB_NEGOC] = 1
      activeConnections[cnx][IAC]    = 0
    else
      printf(f_debug, "Unknown %02x\r\n", Ch)
      activeConnections[cnx][IAC] = 0
    end if
    return
  end if

  if Ch = TNCH_EL then
    puts(f_debug, "TNCH_EL\r\n")
    AddChar(cnx, Ch)
  elsif Ch = TNCH_EC then
    puts(f_debug, "TNCH_EC\r\n")
    AddChar(cnx, Ch)
  elsif Ch = TNCH_AYT then
    puts(f_debug, "TNCH_AYT\r\n")
    AddChar(cnx, Ch)
  elsif Ch = TNCH_IP then
    puts(f_debug, "TNCH_IP\r\n")
    AddChar(cnx, Ch)
  elsif Ch = TNCH_AO then
    puts(f_debug, "TNCH_AO\r\n")
    AddChar(cnx, Ch)
  elsif Ch = TNCH_IAC then
    activeConnections[cnx][IAC] = 1
  else
    AddChar(cnx, Ch)
  end if
end procedure

--------------------------------------------------------------------------------

global function getActiveSockets()
  sequence s
  
  s = {}
  for i = 1 to length(activeConnections) do
    if activeConnections[i][TCP_STATUS] = CONNECTED then
      s = append(s, {activeConnections[i][TCP_SOCKET], POLLIN+POLLERR})
    end if
  end for
  return s
end function

--------------------------------------------------------------------------------

global function findConnection(sequence host, integer port)
  sequence s

  s = commonFields(activeConnections, {TCP_HOST,TCP_PORT}, {host,port}, 1)
  if length(s) then return s[1] else return 0 end if
end function

--------------------------------------------------------------------------------

global procedure telnetClose(integer cnx)
  sequence host
  atom tcpSocket

  if length(activeConnections)=0 then
    logMsg("telnetClose: no active connection available")
    return
  end if
  if activeConnections[cnx][TCP_STATUS] = DISCONNECTED then
    logMsg(sprintf("telnetClose: already disconnected (%d)", cnx))
    return
  end if
  host = activeConnections[cnx][TCP_HOST]
  tcpSocket = activeConnections[cnx][TCP_SOCKET]
  void = eunet_close_socket(tcpSocket)
  activeConnections[cnx][TCP_STATUS] = DISCONNECTED
  activeConnections[cnx][TCP_SOCKET] = 0
  activeSockets = getActiveSockets()
end procedure

--------------------------------------------------------------------------------

procedure telnetReceive(integer cnx)
  object data

  logMsg(sprintf("telnetReceive(%d)", {cnx}))
  if not checkLink(cnx) then return end if
  data = eunet_recv(activeConnections[cnx][TCP_SOCKET], 0)
  if (length(data)=2) and sequence(data[2]) then
    if data[1] = EUNET_ERROR_NODATA then
      -- nothing to de
    elsif data[1] = EUNET_ERROR_NOCONNECTION then
      logMsg(sprintf("telnetReceive: Connection %d lost", {cnx}))
      activeConnections[cnx][TCP_STATUS] = DISCONNECTED
    else
      handleError()
      logMsg(sprintf("telnetReceive: Error while receiving on connection %d", {cnx}))
      telnetClose(cnx)
    end if
  end if
--  printf(f_debug, "bytes received = %d\n", {length(data)})
  analyzeObject(data, sprintf("telnetReceive(%d)", {cnx}), f_debug, 0)
  activeConnections[cnx][LAST_IO] = time()
  for i = 1 to length(data) do ReceiveChar(cnx, data[i]) end for
  data = bufferRead(displayBuffer)
  if sequence (data) and onTheFlowDisplayRoutine then  -- process data with routine onTheFlowDisplayRoutine
    call_proc(onTheFlowDisplayRoutine, {data})
  end if
  bufferClear(displayBuffer)
end procedure

--------------------------------------------------------------------------------

-- has to be launched periodically by a 0.1 s timer
global procedure checkConnections(integer timeout)
  -- get socket state
  object data
  integer cnx

  for i = 1 to length(activeConnections) do
--    logMsg(sprintf("activeConnections[%d][TCP_STATUS] = %d\n", {i, activeConnections[i][TCP_STATUS]}))
--    printf(f_debug, "elapsed time = %d\n", {time() - activeConnections[i][LAST_IO]})
    if activeConnections[i][TCP_STATUS] = CONNECTED then
      if (time() - activeConnections[i][LAST_IO]) >= maxTime then
        void = telnetSend(i, {32,8})  -- SPACE and BACKSPACE
      end if
    end if
  end for
  data = eunet_poll(activeSockets, timeout)
  if atom(data) then
    return -- no data received
  end if
  for i = 1 to length(data) do
    if data[i] > 0 then
      cnx = findRecNumByField(activeConnections, TCP_SOCKET, activeConnections[i][1], 1)
      if not cnx then
        warnError("checkConnections: connection not found", 1)
      end if
      telnetReceive(cnx)
      void = eunet_poll({{activeSockets[i][1],0}},0)
    end if
  end for
end procedure

--------------------------------------------------------------------------------

global function telnetConnect(sequence host, integer port)
  atom tcpSocket
  integer cnx, buffer
  sequence s
  
  tcpSocket = sockConnect(host, port)
  if tcpSocket = 0 then
    logMsg(sprintf("telnetConnect: Could not connect to %s:%d", {host, port}))
    void = eunet_close_socket(tcpSocket)
    return 0
  else
    logMsg(sprintf("telnetConnect: Connected to %s:%d", {host, port}))
    s = commonFields(activeConnections, {TCP_HOST,TCP_PORT}, {host, port}, 1)
    if length(s) then
      cnx = s[1]
      buffer = activeConnections[cnx][BUFFER]
      activeConnections[cnx] = {host, port, tcpSocket, CONNECTED, time(), buffer, 0, #0, "", 0}
      bufferClear(buffer)
    else
      activeConnections = append(activeConnections, {host, port, tcpSocket, CONNECTED, time(), bufferNew(), 0, #0, "", 0})
      cnx = length(activeConnections)
    end if
    activeSockets = getActiveSockets()
  end if
  return cnx
end function

------------------------------------------------------------------------------

global function telnetRead(integer cnx, atom waitTime)
  -- wait waitTime
  -- read buffer
  -- display
  sequence data

--  logMsg(sprintf("telnetRead(%d, %.3f)", {cnx,waitTime}))
  if not checkLink(cnx) then return 0 end if
  puts(f_debug, ".")
  pause(waitTime)
  data = bufferRead(activeConnections[cnx][BUFFER])
  bufferClear(activeConnections[cnx][BUFFER])
  return data
end function

------------------------------------------------------------------------------

global function readUntil(integer cnx, sequence actualInvite, sequence loginInvite, atom waitTime, integer maxTime, integer debugMode)
  -- on lit avec telnetRead aprs waitTime
  -- tant qu'on n'a pas reu ce qui est attendu, on attend encore waitTime et ainsi de suite jusqu' ce qu'on ait atteint maxTime
  -- ds qu'on a reu ce qui est attendu on quitte
  sequence read
  atom t0
  object data

  logMsg(sprintf("readUntil(%d, '%s', '%s', %.3f, %d)", {cnx, actualInvite, loginInvite, waitTime, maxTime}))
  if length(actualInvite)=0 then
    logMsg("readUntil: actual invite missing !")
    return 0
  end if
  t0 = time()
  read = ""
  while 1 do
    data = telnetRead(cnx, waitTime)
    if atom(data) then return 0 end if
    if length(data) then
      read &= data
--      analyzeObject(read, sprintf("readUntil(%d): read", {cnx}), f_debug, 0)
      if match(actualInvite, read) then
        return read
      elsif match(loginInvite, read) then   -- login actualInvite
        exit
      end if
      if debugMode then	-- en mode debug, pas d'invite sans retour chariot
        if time()-t0 > 20 then -- toutes les 20 secondes
          void = telnetSend(cnx, "\n")
          t0 = time()
        end if
      end if
    end if
    if (maxTime != -1) and ( (time()-t0) > maxTime ) then
      logMsg(sprintf("readUntil(%d): maxTime exceeded", {cnx}))
      exit
    end if
  end while
  if length(data) then
    analyzeObject(data, sprintf("readUntil(%d)", {cnx}), f_debug, 0)
  end if
  logMsg(sprintf("readUntil: failed to read '%s'", {actualInvite}))
  return 0
end function

------------------------------------------------------------------------------

global function readAll(integer cnx, atom waitTime, object maxTime)
  -- on lit avec telnetRead aprs waitTime
  -- si on n'a rien reu, on attend encore waitTime et ainsi de suite jusqu' ce
  --   qu'on ait atteint maxTime si c'est un entier ou maxTime[1] (maxWaitTime) si c'est une squence
  -- si on a reu quelque chose, on attend waitTime et ainsi de suite tant qu'on reoit
  --   jusqu' ce qu'on ait atteint maxTime si c'est un entier ou maxTime[2] (maxReadTime) si c'est une squence
  --   on arrte de suite si on ne reoit plus
  sequence read
  atom t0
  object data
  integer flagReceiving

  logMsg(sprintf("readAll(%d, %.3f, %s)", {cnx, waitTime, sprint(maxTime)}))
  t0 = time()
  read = ""
  flagReceiving = 0
  while 1 do
    data = telnetRead(cnx, waitTime)
    if atom(data) then return 0 end if
    if length(data) then
      read &= data
--      analyzeObject(read, sprintf("readAll(%d): read", {cnx}), f_debug, 0)
      if not flagReceiving and sequence(maxTime) then
        t0 = time()
      end if
      flagReceiving = 1
      if atom(maxTime) then
        if (maxTime != -1) and ( (time()-t0) > maxTime ) then
          logMsg(sprintf("readAll(%d): maxTime exceeded", cnx))
          exit
        end if
      else
        if (time()-t0) > maxTime[2] then
          logMsg(sprintf("readAll(%d): maxReadTime exceeded", cnx))
          exit
        end if
      end if
    else
      if not flagReceiving then
        if atom(maxTime) then
          if (maxTime != -1) and ( (time()-t0) > maxTime ) then
            logMsg(sprintf("readAll(%d): maxTime exceeded", cnx))
            exit
          end if
        else
          if (time()-t0) > maxTime[1] then
            logMsg(sprintf("readAll(%d): maxWaitTime exceeded", cnx))
            exit
          end if
        end if
      else  --  on a reu mais on ne reoit plus: on arrte de suite
        exit
      end if      
    end if
  end while
  return read
end function

------------------------------------------------------------------------------

global function sendReceive(integer cnx, sequence cmd, sequence actualInvite, sequence loginInvite, atom waitTime, integer maxTime)
  logMsg(sprintf("sendReceive(%d, '%s', '%s', '%s', %.3f, %d)",{cnx, showPrintable(cmd), actualInvite, loginInvite, waitTime, maxTime}))
  if telnetSend(cnx, cmd) < 0 then return 0 end if
  return readUntil(cnx, actualInvite, loginInvite, waitTime, maxTime, 1)
end function

