-- myTypes.e
-- v1.1
-- 2016-09-26
-- common types

-- type positive(atom x)
-- checks if x >=0

-- type is_space(integer n)
-- checks if an char is a space, tab, vtab, cr, lf of ff

-- type is_alpha(integer n)
-- checks if a char is within [a-z] or [A-Z]

-- type is_digit(integer n)
-- test if a char is within [0-9]

-- type is_string(object x)
-- checks if a sequence is a string

-- type is_number(sequence x)
-- checks if a sequence represents a number

-- type is_date(sequence x)
-- checks if a sequence represents a date
-- [.]./[.]./[..]..

-- type is_hex(sequence x)
-- checks if a sequence represents a hexadecimal number

-- type is_identifier(sequence s)
-- checks if a sequence represents an identifier
-- begins with '_' or  [a-z] or [A-Z]
-- contains '_' or  [a-z] or [A-Z] or [0-9]

------------------------------------------------------------------------------

global constant
  ON = 1,
  OFF = 0,
  SUCCESS = 1,
  FAIL = 0

global constant
  DIGITS = "0123456789ABCDEF"

global object
  void    -- used when return value doesn't care

------------------------------------------------------------------------------

-- checks if x >=0
global type positive(atom x)
  return x>=0
end type

--------------------------------------------------------------------------------

-- checks if an char is a space, tab, vtab, cr, lf of ff
global type is_space(integer n)
  return find(n, {32, 10, 9, 11, 12, 13})
end type

--------------------------------------------------------------------------------

-- checks if a char is within [a-z] or [A-Z]
global type is_alpha(integer n)
  return ((n >= 'A') and (n <= 'Z')) or
         ((n >= 'a') and (n <= 'z'))
end type

--------------------------------------------------------------------------------

-- test if a char is within [0-9]
global type is_digit(integer n)
  return (n >= '0') and (n <= '9')
end type

------------------------------------------------------------------------------

-- checks if a sequence is a string
global type is_string(object x)
  if atom(x) then return 0 end if
  for i = 1 to length(x) do
    if sequence(x[i]) then return 0 end if
    if (x[i] < ' ') and (x[i] != 9) and (x[i] != 13) and (x[i] != 10) then return 0 end if
    if x[i]> 254 then return 0 end if
  end for
  return 1
end type

------------------------------------------------------------------------------

-- checks if a sequence represents a number
global type is_number(sequence s)
  integer lg

  lg = length(s)
  if lg = 0 then return 0 end if
  for i = 1 to lg do
    if sequence(s[i]) then
      return 0
    else
      if not is_digit(s[i]) and not find(s[i], ",.-+") then
        return 0
      end if
    end if
  end for
  return 1
end type

------------------------------------------------------------------------------

-- checks if a sequence represents a date
-- [.]./[.]./[..]..
global type is_date(sequence x)
  integer n, n2

  n = length(x) if (n < 6) or (n > 10) then return 0 end if
  for i = 1 to length(x) do
    if sequence(x[i]) then return 0 end if
    if not is_digit(x[i]) and (x[i] != '/') then return 0 end if
    n = find('/', x) if (n != 2) and (n != 3) then return 0 end if
    n2 = find('/', x[n+1..$]) if (n2 != 2) and (n2 != 3) then return 0 end if
  end for
  return 1
end type

------------------------------------------------------------------------------

-- checks if a sequence represents a hexadecimal number
global type is_hex(sequence x)
  integer lg
  sequence s

  s = x
  lg = length(s)
  if lg <= 1 then return 0 end if         -- min 2 cars (#n)
  if s[1] != '#' then return 0 end if
  for i = 2 to lg do
    if sequence(s[i]) then
      return 0
    else
      if not find(s[i], DIGITS) then return 0 end if
    end if
  end for
  return 1
end type

--------------------------------------------------------------------------------

-- checks if a sequence represents an identifier
-- begins with '_' or  [a-z] or [A-Z]
-- contains '_' or  [a-z] or [A-Z] or [0-9]
global type is_identifier(sequence s)
  integer lg

  lg = length(s)
  if lg = 0 then return 0 end if
  if not is_alpha(s[1]) and (s[1] != '_') then
    return 0
  end if
  for i = 2 to lg do
    if sequence(s[i]) then
      return 0
    else
      if not is_alpha(s[i]) and (s[i] != '_') and not is_digit(s[i]) then
        return 0
      end if
    end if
  end for
  return 1
end type

